#!/usr/bin/python

import numpy as np

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pyrodeo

def test_sph():
    sim = pyrodeo.Simulation.from_geom('sph',
                                       dimensions=[128, 1, 32],
                                       domain=([0.4, 2.5],
                                               [],
                                               [0.5*np.pi-0.15, 0.5*np.pi]),
                                       log_radial=True)

    # Sound speed constant H/r = 0.05
    sim.state.soundspeed = 0.05*sim.state.soundspeed
    sim.param.boundaries[0] = 'reflect'
    sim.param.boundaries[1] = 'periodic'
    sim.param.boundaries[2] = 'reflect'

    # Simulate up to 0.01 orbits
    sim.evolve([0.01*2.0*np.pi], new_file=True)



# -*- coding: utf-8 -*-
"""Defines a generic conservation law solver.
"""

from __future__ import print_function

import numpy as np

class ClawSolver:
    """Generic conservation law solver.

    Note:
        Serves as a base class to construct various solvers. Can not be used on its own.

    The following method is available:

    """
    def limiter(self, a, b, sb):
        """Limiter function to limit slopes/fluxes.

        This limiter function can, based on the parameter sb, emulate total variation diminishing limiters from minmod (sb = 1) to superbee (sb = 2).

        Args:
            a (ndarray): First slope to compare.
            b (ndarray): Second slope to compare.
            sb (float): Limiter parameter. Should be >= 1 (minmod, most diffusive limiter) and <= 2 (superbee, least diffusive limiter).

        Returns:
            ndarray: Limited slopes.

        """
        return (np.maximum(0.0*a,
                           np.minimum(sb*a,
                                      np.maximum(b, np.minimum(a, sb*b)))) +
                np.minimum(0.0*a,
                           np.maximum(sb*a,
                                      np.minimum(b, np.maximum(a, sb*b)))))



# -*- coding: utf-8 -*-
"""Defines the Roe class defining the Roe solver.
"""

from __future__ import print_function

import numpy as np
from pyrodeo.claw_solver import ClawSolver

class Roe(ClawSolver):
    """Construct class for the Roe solver.

    Constructor sets two basic attributes (sb and min_dens), after which a time step can be taken through the :meth:`.step` method.

    Args:
        flux_limiter (float): Flux limiter parameter. Should be >= 1 (minmod, most diffusive limiter) and <= 2 (superbee, least diffusive limiter).
        min_dens (float): Minimum density when to switch to HLL to preserve positivity.

    The following attributes and methods are available:

    Attributes:
        sb (float): Flux limiter parameter. Should be >= 1 (minmod, most diffusive limiter) and <= 2 (superbee, least diffusive limiter).
        min_dens (float): Minimum density when to switch to HLL to preserve positivity.
    """
    def __init__(self, flux_limiter, min_dens):
        self.sb = flux_limiter
        self.min_dens = min_dens

    def limit_flux(self, dens, dens_left, f1dens, f2dens, dtdx):
        """Limit second order flux to preserve positivity

        Calculate the maximum contribution of the second order flux in order for the density to remain positive.

        Args:
            dens (ndarray): Current density.
            dens_left (ndarray): Density in the cell to the left
            f1dens (ndarray): First order mass flux.
            f2dens (ndarray): Second order mass flux.
            dtdx (float): Time step / space step

        Returns:
            ndarray: Array with values >= 0 and <= 1 specifying the maximum contribution of second order flux for the density to remain positive.

        """
        thetap = 0.0*dens + 1.0
        rhop = 0.5*(dens_left - 2.0*dtdx*f2dens)
        epsd = np.minimum(0.0*dens + self.min_dens,
                          np.minimum(dens_left, dens))
        sel = np.where(rhop < epsd)
        rho1 = 0.5*(dens_left[sel] - 2.0*dtdx*f1dens[sel])
        if len(rho1) >= 1:
            if np.min(rho1) < 0.0:
                dens[sel] = -1.0*dens[sel]
        thetap[sel] = np.maximum(0.0*rho1,
                                 (rho1 - epsd[sel])/(rho1 - rhop[sel]))
        if np.max(thetap) > 1.0:
            dens = -1.0*dens
        thetam = 0.0*dens + 1.0
        rhom = 0.5*(dens + 2.0*dtdx*f2dens)
        sel = np.where(rhom < epsd)
        rho1 = 0.5*(dens[sel] + 2.0*dtdx*f1dens[sel])
        if len(rho1) >= 1:
            if np.min(rho1) < 0.0:
                dens[sel] = -1.0*dens[sel]
        thetam[sel] = np.maximum(0.0*rho1,
                                 (rho1 - epsd[sel])/(rho1 - rhom[sel]))
        if np.max(thetam) > 1.0:
            dens = -1.0*dens

        return np.minimum(thetap, thetam)

    def step(self, dt, dx, state, source, bc):
        """Update state for a single time step.

        Args:
            dt (float): Time step.
            dx (float): Space step.
            state (:class:`.State`): Current :class:`.State`, will be updated.
            source (ndarray): Geometric source terms, must have same shape as state.dens.
            bc (str, str): Boundary conditions (in and out): 'periodic', 'symmetric', or 'closed' (other boundary conditions are dealt with elsewhere).

        """
        # Set periodic boundaries
        if bc[0] == 'periodic':
            state.dens[:2,:,:] = state.dens[-4:-2,:,:]
            state.velx[:2,:,:] = state.velx[-4:-2,:,:]
            state.vely[:2,:,:] = state.vely[-4:-2,:,:]
            state.velz[:2,:,:] = state.velz[-4:-2,:,:]

            state.dens[-2:,:,:] = state.dens[2:4,:,:]
            state.velx[-2:,:,:] = state.velx[2:4,:,:]
            state.vely[-2:,:,:] = state.vely[2:4,:,:]
            state.velz[-2:,:,:] = state.velz[2:4,:,:]

        # Symmetry boundary
        if bc[0] == 'symmetric':
            state.dens[:2,:,:] = state.dens[3:1:-1,:,:]
            state.velx[:2,:,:] = -state.velx[3:1:-1,:,:]
            state.vely[:2,:,:] = state.vely[3:1:-1,:,:]
            state.velz[:2,:,:] = state.velz[3:1:-1,:,:]
        if bc[1] == 'symmetric':
            state.dens[-2:,:,:] = state.dens[-3:-5:-1,:,:]
            state.velx[-2:,:,:] = -state.velx[-3:-5:-1,:,:]
            state.vely[-2:,:,:] = state.vely[-3:-5:-1,:,:]
            state.velz[-2:,:,:] = state.velz[-3:-5:-1,:,:]

        dens_left = np.roll(state.dens, 1, axis=0)
        velx_left = np.roll(state.velx, 1, axis=0)
        vely_left = np.roll(state.vely, 1, axis=0)
        velz_left = np.roll(state.velz, 1, axis=0)

        cs_left = np.roll(state.soundspeed, 1, axis = 0)
        # Average sound speed 0.5*((i) + (i-1))
        cs_left = np.roll(state.soundspeed, 1, axis = 0)
        cs = 0.5*(state.soundspeed + cs_left)

        dtdx = dt/dx

        # Parameter vector
        z0 = np.sqrt(state.dens)
        z1 = z0*state.velx
        z2 = z0*state.vely
        z3 = z0*state.velz

        # Average parameter vector: 0.5*((i) + (i-1))
        z0 = 0.5*(z0 + np.roll(z0, 1, axis=0))
        z1 = 0.5*(z1 + np.roll(z1, 1, axis=0))
        z2 = 0.5*(z2 + np.roll(z2, 1, axis=0))
        z3 = 0.5*(z3 + np.roll(z3, 1, axis=0))

        # Roe averages
        u = z1/z0
        v = z2/z0
        w = z3/z0

        # HLL signal speeds
        bm = np.minimum(0.0*u, np.minimum(u - cs,
                                          velx_left - cs_left))
        bp = np.maximum(0.0*u, np.maximum(u + cs,
                                          state.velx + state.soundspeed))

        # Cell-centred fluxes
        fdens = state.dens*state.velx
        fmomx = fdens*state.velx + \
          state.soundspeed*state.soundspeed*state.dens
        fmomy = fdens*state.vely
        fmomz = fdens*state.velz

        # Left and right fluxes (note stationary extrapolation)
        fldens = np.roll(fdens, 1, axis=0)
        flmomx = np.roll(fmomx + 0.5*dx*source, 1, axis=0)
        flmomy = np.roll(fmomy, 1, axis=0)
        flmomz = np.roll(fmomz, 1, axis=0)

        frdens = fdens
        frmomx = fmomx - 0.5*dx*source
        frmomy = fmomy
        frmomz = fmomz

        # Flux difference: (i) - (i-1)
        dfdens =  frdens - fldens
        dfmomx =  frmomx - flmomx
        dfmomy =  frmomy - flmomy
        dfmomz =  frmomz - flmomz

        # State difference: (i) - (i-1)
        ddens = state.dens - dens_left
        dmomx = state.dens*state.velx - dens_left*velx_left
        dmomy = state.dens*state.vely - dens_left*vely_left
        dmomz = state.dens*state.velz - dens_left*velz_left

        # HLL flux
        f1dens = (bp*fldens - bm*frdens + bp*bm*ddens)/(bp - bm);
        f1momx = (bp*flmomx - bm*frmomx + bp*bm*dmomx)/(bp - bm);
        f1momy = (bp*flmomy - bm*frmomy + bp*bm*dmomy)/(bp - bm);
        f1momz = (bp*flmomz - bm*frmomz + bp*bm*dmomz)/(bp - bm);

        # Projection coefficients
        b1 =-0.5*(dfmomx - (u + cs)*dfdens)/cs
        b2 = 0.5*(dfmomx - (u - cs)*dfdens)/cs
        b3 = dfmomy - v*dfdens;
        b4 = dfmomz - w*dfdens;

        b1 = b1/(u - cs + 1.0e-10)
        b2 = b2/(u + cs + 1.0e-10)
        b3 = b3/(u + 1.0e-10)
        b4 = b4/(u + 1.0e-10)

        # Upwind projection coefficients
        b1u = -0.5*(np.sign(u - cs) - 1.0)*np.roll(b1, -1, axis=0) + \
          0.5*(np.sign(u - cs) + 1.0)*np.roll(b1, 1, axis=0)
        b2u = -0.5*(np.sign(u + cs) - 1.0)*np.roll(b2, -1, axis=0) + \
          0.5*(np.sign(u + cs) + 1.0)*np.roll(b2, 1, axis=0)
        b3u = -0.5*(np.sign(u) - 1.0)*np.roll(b3, -1, axis=0) + \
          0.5*(np.sign(u) + 1.0)*np.roll(b3, 1, axis=0)
        b4u = -0.5*(np.sign(u) - 1.0)*np.roll(b4, -1, axis=0) + \
          0.5*(np.sign(u) + 1.0)*np.roll(b4, 1, axis=0)

        # Flux limiter
        fl1 = self.limiter(b1, b1u, self.sb)
        fl2 = self.limiter(b2, b2u, self.sb)
        fl3 = self.limiter(b3, b3u, self.sb)
        fl4 = self.limiter(b4, b4u, self.sb)

        # Second order projection coefficients
        b1 = -np.abs(u - cs)*b1 + fl1*(np.abs(u - cs) - dtdx*(u - cs)*(u - cs))
        b2 = -np.abs(u + cs)*b2 + fl2*(np.abs(u + cs) - dtdx*(u + cs)*(u + cs))
        b3 = -np.abs(u)*b3 + fl3*(np.abs(u) - dtdx*u*u)
        b4 = -np.abs(u)*b4 + fl4*(np.abs(u) - dtdx*u*u)

        # Interface flux (i - 1/2)
        f2dens = 0.5*(fldens + frdens + b1 + b2)
        f2momx = 0.5*(flmomx + frmomx + (u - cs)*b1 + (u + cs)*b2)
        f2momy = 0.5*(flmomy + frmomy + v*(b1 + b2) + b3)
        f2momz = 0.5*(flmomz + frmomz + w*(b1 + b2) + b3)

        # Limit flux to maintain positivity
        theta = self.limit_flux(state.dens, dens_left, f1dens, f2dens, dtdx)

        # Limited fluxes
        fdens = theta*f2dens - (theta - 1.0)*f1dens
        fmomx = theta*f2momx - (theta - 1.0)*f1momx
        fmomy = theta*f2momy - (theta - 1.0)*f1momy
        fmomz = theta*f2momz - (theta - 1.0)*f1momz

        # Adjust boundary fluxes for closed boundaries
        if bc[0] == 'closed':
            c = state.soundspeed[2,:,:]
            fdens[2,:,:] = 0.0
            fmomx[2,:,:] = c*c*state.dens[2,:,:] - 0.5*dx*source[2,:,:]
            fmomy[2,:,:] = 0.0
            fmomz[2,:,:] = 0.0
        if bc[1] == 'closed':
            c = state.soundspeed[-3,:,:]
            fdens[-2,:,:] = 0.0
            fmomx[-2,:,:] = c*c*state.dens[-3,:,:] + 0.5*dx*source[-3,:,:]
            fmomy[-2,:,:] = 0.0
            fmomz[-2,:,:] = 0.0

        if bc[0] == 'symmetric':
            fdens[2,:,:] = 0.0
            fmomy[2,:,:] = 0.0
            fmomz[2,:,:] = 0.0
        if bc[1] == 'symmetric':
            fdens[-2,:,:] = 0.0
            fmomy[-2,:,:] = 0.0
            fmomz[-2,:,:] = 0.0

        # Interface flux difference: (i+1) - (i)
        dfdens = np.roll(fdens, -1, axis=0) - fdens
        dfmomx = np.roll(fmomx, -1, axis=0) - fmomx
        dfmomy = np.roll(fmomy, -1, axis=0) - fmomy
        dfmomz = np.roll(fmomz, -1, axis=0) - fmomz

        # Change in density and momenta (ignoring ghost cells)
        ddens = -dtdx*dfdens*state.no_ghost
        dmomx = (dt*source - dtdx*dfmomx)*state.no_ghost
        dmomy = -dtdx*dfmomy*state.no_ghost
        dmomz = -dtdx*dfmomz*state.no_ghost

        state.dens += ddens
        state.velx += (dmomx - ddens*state.velx)/state.dens
        state.vely += (dmomy - ddens*state.vely)/state.dens
        state.velz += (dmomz - ddens*state.velz)/state.dens



#!/usr/bin/python

import numpy as np

from pyrodeo.state import State
from pyrodeo.coords import Coordinates
import pyrodeo.linear_advection as pr

def linear_advection_error():
    coords = Coordinates.from_dims(dimensions=(100, 1, 1),
                                   domain=([-0.5, 0.5], [], []))
    state = State.from_dims((104,1,1))

    state.dens += np.cos(2.0*np.pi*coords.x)
    advection_velocity = 0.0*coords.x + 1.0

    la = pr.LinearAdvection(advection_velocity, 1.0)

    maxn = 17
    for n in range(maxn):
        la.step(1.0/maxn, coords.dxyz[0], state)

    dens_error = np.sum(np.abs(state.dens - 1.0 -
                               np.cos(2.0*np.pi*coords.x)))

    return dens_error

def test_linear_advection():
    assert linear_advection_error() < 0.02



# -*- coding: utf-8 -*-
"""Defines a linear advection solver in a periodic domain.
"""

from __future__ import print_function

import numpy as np
from pyrodeo.claw_solver import ClawSolver

class LinearAdvection(ClawSolver):
    """Linear advection solver in periodic domain.

    Args:
        advection_velocity (ndarray): Advection velocity. Must have the same shape as :class:`.State` members (density, velocity) to be advected and must be uniform in the first dimension (x).
        limiter_parameter (float): Parameter setting the limiter function. Should be >= 1 (minmod limiter, most diffusive) and <= 2 (superbee limiter, least diffusive).

    Note:
        The linear advection equation is solved on a periodic x domain. The advection velocity can not depend on x. If advecting over y, the state will have to be transposed first.

    The following public attributes and methods are available:

    Attributes:
        advection_velocity (ndarray): Advection velocity. Must have the same shape as :class:`.State` members (density, velocity) to be advected and must be uniform in the first dimension (x).
        sb (float): Parameter setting the limiter function. Should be >= 1 (minmod limiter, most diffusive) and <= 2 (superbee limiter, least diffusive).

    """
    def __init__(self, advection_velocity, limiter_parameter):
        if (limiter_parameter < 1.0 or limiter_parameter > 2.0):
            raise ValueError('Expecting limiter_parameter >= 1 and <= 2 to be TVD')

        self.advection_velocity = advection_velocity
        self.sb = limiter_parameter

    def step(self, dt, dx, state):
        """Perform one time step dt.

        Evolve the linear advection equation over a time step dt, updating the state. The domain can be multidimensional, but evolution is with respect to x. There is no restriction on the magnitude of dt.

        Args:
            dt (float): Time step to take.
            dx (float): Step size in x.
            state (State): :class:`.State` containing density and velocity

        """
        # Length of state in x and y
        nx = len(state.dens[:,0,0])
        ny = len(state.dens[0,:,0])

        # Case of single cell in x
        if nx == 1:
            return

        # Work with momenta
        state.velx = state.velx*state.dens
        state.vely = state.vely*state.dens
        state.velz = state.velz*state.dens

        # Advection velocity
        u = self.advection_velocity[2:nx-2,:,:]

        # Distance to shift solution in x
        shiftx = dt*u

        # Number of integer cells to shift
        Nshift = np.around(shiftx/dx)
        # Remaining u*dt/dx
        udtdx = shiftx/dx - Nshift

        # Views without ghost cells
        dens = state.dens[2:nx-2,:,:]
        momx = state.velx[2:nx-2,:,:]
        momy = state.vely[2:nx-2,:,:]
        momz = state.velz[2:nx-2,:,:]

        # Limited slopes for 2nd order correction
        slopedens = self.limiter(dens - np.roll(dens, 1, axis=0),
                                 np.roll(dens, -1, axis=0) - dens, self.sb)
        slopemomx = self.limiter(momx - np.roll(momx, 1, axis=0),
                                 np.roll(momx, -1, axis=0) - momx, self.sb)
        slopemomy = self.limiter(momy - np.roll(momy, 1, axis=0),
                                 np.roll(momy, -1, axis=0) - momy, self.sb)
        slopemomz = self.limiter(momz - np.roll(momz, 1, axis=0),
                                 np.roll(momz, -1, axis=0) - momz, self.sb)

        # Upwind state
        s = np.sign(udtdx)
        densu = -0.5*(s - 1.0)*np.roll(dens, -1, axis=0) + \
            0.5*(s + 1.0)*np.roll(dens, 1, axis=0)
        momxu = -0.5*(s - 1.0)*np.roll(momx, -1, axis=0) + \
            0.5*(s + 1.0)*np.roll(momx, 1, axis=0)
        momyu = -0.5*(s - 1.0)*np.roll(momy, -1, axis=0) + \
            0.5*(s + 1.0)*np.roll(momy, 1, axis=0)
        momzu = -0.5*(s - 1.0)*np.roll(momz, -1, axis=0) + \
            0.5*(s + 1.0)*np.roll(momz, 1, axis=0)

        # Upwind slopes
        slopedensu = -0.5*(s - 1.0)*np.roll(slopedens, -1, axis=0) + \
            0.5*(s + 1.0)*np.roll(slopedens, 1, axis=0)
        slopemomxu = -0.5*(s - 1.0)*np.roll(slopemomx, -1, axis=0) + \
            0.5*(s + 1.0)*np.roll(slopemomx, 1, axis=0)
        slopemomyu = -0.5*(s - 1.0)*np.roll(slopemomy, -1, axis=0) + \
            0.5*(s + 1.0)*np.roll(slopemomy, 1, axis=0)
        slopemomzu = -0.5*(s - 1.0)*np.roll(slopemomz, -1, axis=0) + \
            0.5*(s + 1.0)*np.roll(slopemomz, 1, axis=0)

        # Update state
        a = np.abs(udtdx);
        dens += (-a*(dens - densu) - \
                 0.5*udtdx*(1.0 - a)*(slopedens - slopedensu))
        momx += (-a*(momx - momxu) - \
                 0.5*udtdx*(1.0 - a)*(slopemomx - slopemomxu))
        momy += (-a*(momy - momyu) - \
                 0.5*udtdx*(1.0 - a)*(slopemomy - slopemomyu))
        momz += (-a*(momz - momzu) - \
                 0.5*udtdx*(1.0 - a)*(slopemomz - slopemomzu))

        # Shift integer number of cells
        for i in range(ny):
            n = int(Nshift[0,i,0])
            dens[:,i,:] = np.roll(dens[:,i,:], n, axis=0)
            momx[:,i,:] = np.roll(momx[:,i,:], n, axis=0)
            momy[:,i,:] = np.roll(momy[:,i,:], n, axis=0)
            momz[:,i,:] = np.roll(momz[:,i,:], n, axis=0)

        # Switch back to velocities
        state.velx /= state.dens
        state.vely /= state.dens
        state.velz /= state.dens



#!/usr/bin/python

import numpy as np

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pyrodeo.simulation as prs

def test_cartesian():
    # Create simulation with default resolution and domain
    sim = prs.Simulation.from_geom('cart')

    # The basic state will have density and sound speed unity everywhere,
    # and the velocity will be zero. In order to create a simple shock tube,
    # now set density to 1/10 for x > 0.
    sel = np.where(sim.coords.x > 0.0)
    sim.state.dens[sel] = 0.1

    # Evolve until t = 0.25
    sim.evolve([0.25], new_file=True)



#!/usr/bin/python

import numpy as np

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pyrodeo

# Extra source terms: planet gravity
def planet_source(t, dt, coords, state, planetParam):
    # Mass ratio planet/star
    mp = planetParam[0]
    # Softening length planet potential
    eps = planetParam[1]

    # Coordinates
    r = coords.x
    p = coords.y

    # Planet coordinates
    rp = 1.0
    pp = 0.0

    # Distance to the planet
    dist = np.sqrt(r*r + rp*rp - 2.0*r*rp*np.cos(p - pp) + eps*eps)

    # Potential gradient
    dpotdr = mp*(r - rp*np.cos(p - pp))/(dist*dist*dist)
    dpotdp = mp*r*rp*np.sin(p - pp)/(dist*dist*dist)

    # Indirect term
    dpotdr += mp*np.cos(p - pp)/(rp*rp)
    dpotdp -= mp*r*np.sin(p - pp)/(rp*rp)

    # Resulting source term
    source_velx = -dpotdr
    source_vely = -dpotdp/(r*r)

    # Damping boundary conditions
    Rin = 100.0*(r - 0.5)*(r - 0.5)
    Rin[np.where(r > 0.5)] = 0.0
    Rout = (r - 2.1)*(r - 2.1)/(0.4*0.4)
    Rout[np.where(r < 2.1)] = 0.0
    R = (Rin + Rout)*np.power(r, -1.5)

    # Damp towards initial state
    source_dens = -(state.dens - 1.0)*R
    source_velx -= state.velx*R
    source_vely -= state.vely*R

    # Integrate extra source terms
    state.dens += dt*source_dens*state.no_ghost
    state.velx += dt*source_velx*state.no_ghost
    state.vely += dt*source_vely*state.no_ghost

def test_cyl():
    sim = pyrodeo.Simulation.from_geom('cyl',
                                       dimensions=[128, 384, 1],
                                       domain=([0.4, 2.5],[-np.pi, np.pi],[]))

    # Sound speed constant H/r = 0.05
    sim.state.soundspeed = 0.05*sim.state.soundspeed/np.sqrt(sim.coords.x)
    sim.param.boundaries[0] = 'reflect'
    sim.param.boundaries[1] = 'periodic'

    # Simulate a Jupiter planet up to 0.01 orbits
    sim.evolve([0.01*2.0*np.pi],
               planet_source, (0.001, 0.6*0.05), new_file=True)

    # Axisymmetric, logarithmic (r,z)
    sim = pyrodeo.Simulation.from_geom('cyl',
                                       dimensions=[128, 1, 32],
                                       domain=([0.4, 2.5],[],[-0.1,0.1]),
                                       log_radial=True)

    sim.evolve([0.01*2.0*np.pi], new_file=True)



# -*- coding: utf-8 -*-
"""Coordinate class used in pyrodeo.

The Coordinate class hold the x, y and z coordinates as 3D ndarrays. In addition, it holds the step size in x, y and z and the size of the grid in x, y and z.
"""

from __future__ import print_function

import numpy as np

class Coordinates(object):
    """Class containing coordinates used in pyrodeo.

    Args:
        x (ndarray): 2D ndarray containing x coordinates
        y (ndarray): 2D ndarray containing y coordinates
        z (ndarray): 2D ndarray containing z coordinates
        log_radial (:obj:`bool`, optional): Flag whether x is a logarithmic radial coordinate.

    Note:
        The validity of the arrays is not checked! To be used in a simulation, they should have the same shape, with x[:,j,k] containing the x coordinates for all j,k, y[i,:,k] containing the y coordinates for all i,k and z[i,j,:] containing the z coordinates for all i,j. In addition, x, y and z should have a constant step size.

    The following public attributes are available:

    Attributes:
        x (ndarray): 2D ndarray containing x coordinates
        y (ndarray): 2D ndarray containing y coordinates
        z (ndarray): 3D ndarray containing z coordinates
        dimensions ([int, int, int]): grid dimensions in the x, y and z direction
        dxyz ([float, float, float]): step size in the x, y and z direction
        log_radial (:obj:`bool`, optional): Flag whether x is a logarithmic radial coordinate.

    """

    def __init__(self, x, y, z, log_radial=False):
        self.dimensions = (len(x[:,0,0]), len(y[0,:,0]), len(z[0,0,:]))
        self.dxyz = [0.0, 0.0, 0.0]
        self.log_radial = log_radial

        if self.dimensions[0] > 1:
            self.dxyz[0] = x[1,0,0] - x[0,0,0]
            if self.log_radial == True:
                self.dxyz[0] = np.log(x[1,0,0]/x[0,0,0])
        if self.dimensions[1] > 1:
            self.dxyz[1] = y[0,1,0] - y[0,0,0]
        if self.dimensions[2] > 1:
            self.dxyz[2] = z[0,0,1] - z[0,0,0]

        self.x = x
        self.y = y
        self.z = z

    @classmethod
    def from_1d(cls, x, y, z, log_radial=False):
        """Initialize from 1D arrays.

        Build coordinates from existing 1D ndarrays (should include ghost cells). Calculate dimensions and step sizes.

        Args:
            x (ndarray): 1D ndarray containing x coordinates
            y (ndarray): 1D ndarray containing y coordinates
            z (ndarray): 1D ndarray containing z coordinates
            log_radial (:obj:`bool`, optional): Flag whether x is a logarithmic radial coordinate.

        Note:
            The validity of arrays x, y and z is not checked. They should have constant step size.

        """
        x, y, z = np.meshgrid(x, y, z, indexing='ij')
        return cls(x, y, z, log_radial=log_radial)

    @classmethod
    def from_dims(cls,
                  dimensions=(100, 1, 1),
                  domain=([-0.5, 0.5], [], []),
                  log_radial=False):
        """Initialize from dimensions and domain size.

        Build coordinates given the dimensions of the grid and the size of the domain. Some basic checks are performed to ensure the resulting coordinates are valid.

        Args:
            dimensions (:obj:`(int,int,int)`, optional): Dimensions of the grid
            domain (:obj:`([float,float],[float,float],[float,float])`, optional): Domain boundaries in x, y and z
            log_radial (:obj:`bool`, optional): Flag whether x will be a logarithmic radial coordinate.
        """
        # Check if dimensions valid
        if len(dimensions) != 3:
            raise TypeError('Expected dimensions to have three elements')
        if (dimensions[1] != 1 and len(domain[1]) != 2):
            raise TypeError('Invalid y domain')
        if (dimensions[2] != 1 and len(domain[2]) != 2):
            raise TypeError('Invalid z domain')
        if np.max(dimensions) <= 1:
            raise ValueError('Need one dimension to be larger than 1')

        # Check if domain is valid
        if len(domain) != 3:
            raise TypeError('Expected domain to have three elements')

        # Step sizes
        dxy = [0.0, 0.0, 0.0]

        # Add x if necessary
        if dimensions[0] > 1:
            if domain[0][1] <= domain[0][0]:
                raise ValueError('Invalid x domain')

            # Convert boundaries to log
            if log_radial is True:
                domain[0][0] = np.log(domain[0][0])
                domain[0][1] = np.log(domain[0][1])

            dxy[0] = (domain[0][1] - domain[0][0])/np.float(dimensions[0])

            # x coordinates, including two ghost cells on either side
            x = np.linspace(domain[0][0] - 1.5*dxy[0],
                            domain[0][1] + 2.5*dxy[0],
                            dimensions[0] + 4,
                            endpoint=False)

            # Make x normal radius (but unevenly spaced!)
            if log_radial is True:
                x = np.exp(x)

        else:
            x = [0.0]

        # Add y if necessary
        if dimensions[1] > 1:
            if domain[1][1] <= domain[1][0]:
                raise ValueError('Invalid y domain')
            dxy[1] = (domain[1][1] - domain[1][0])/np.float(dimensions[1])
            y = np.linspace(domain[1][0] - 1.5*dxy[1],
                            domain[1][1] + 2.5*dxy[1],
                            dimensions[1] + 4,
                            endpoint=False)
        else:
            y = [0.0]

        # Add z if necessary
        if dimensions[2] > 1:
            if domain[2][1] <= domain[2][0]:
                raise ValueError('Invalid z domain')
            dxy[2] = (domain[2][1] - domain[2][0])/np.float(dimensions[2])
            z = np.linspace(domain[2][0] - 1.5*dxy[2],
                            domain[2][1] + 2.5*dxy[2],
                            dimensions[2] + 4,
                            endpoint=False)
        else:
            z = [0.0]

        return cls.from_1d(x, y, z, log_radial=log_radial)

    def __str__(self):
        """Show information about coordinates."""
        return 'Coordinate object, part of pyrodeo. \n \
    Grid dimensions (nx, ny, nz): {} \n \
    Domain size x (including ghost zones): {}, {} \n \
    Domain size y (including ghost zones): {}, {} \n \
    Domain size z (including ghost zones): {}, {}\n'.format(np.shape(self.x),
                                                            np.min(self.x),
                                                            np.max(self.x),
                                                            np.min(self.y),
                                                            np.max(self.y),
                                                            np.min(self.z),
                                                            np.max(self.z))



#!/usr/bin/python

import numpy as np

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pyrodeo

def test_sheet():
    # Domain half-width in x and y
    Lx = 2.0
    Ly = 20.0

    # Create simulation, setting grid dimensions and domain
    sim = pyrodeo.Simulation.from_geom('sheet',
                                       dimensions=[32, 64, 1],
                                       domain=([-Lx, Lx], [-Ly, Ly], []))
    sim.param.boundaries[0] = 'shear periodic'
    sim.param.boundaries[1] = 'periodic'

    # Density profile: single maximum in middle of domain
    sim.state.dens = 0.5*np.cos(np.pi*sim.coords.x/Lx) + 1.0
    # Equilibrium vy to compensate for pressure gradient
    sim.state.vely = \
      -0.25*np.pi*np.sin(np.pi*sim.coords.x/Lx)/(sim.state.dens*Lx)
    # Add some noise to seed instability
    sim.state.dens += 0.01*np.random.random_sample(np.shape(sim.state.dens))

    # Evolve until t = 1.0
    sim.evolve([1.0], new_file=True)

    # Try restoring
    sim_restore = pyrodeo.Simulation.from_checkpoint('./')

    # Test vertical
    sim = pyrodeo.Simulation.from_geom('sheet',
                                       dimensions=[1, 1, 128],
                                       domain=([], [], [-2.0,2.0]))

    sim.param.boundaries[0] = 'shear periodic'
    sim.param.boundaries[1] = 'periodic'
    sim.param.boundaries[2] = 'reflect'

    sim.evolve([1.0], new_file=True)

    # Restore and continue
    sim_restore = pyrodeo.Simulation.from_checkpoint('./')
    sim_restore.evolve([2.0])
